@@ -106,6 +106,7 @@
 #include <string.h>
 #include <assert.h>
 #include <stdarg.h>
+#include <stdio.h>
 
 #if defined(__cplusplus)
 extern "C" {
@@ -115,7 +116,7 @@ extern "C" {
  * {{{ Platform-specific definitions
  */
 
-/** \cond 0 **/
+/** \cond false **/
 
 #if defined(__CC_ARM)         /* set the alignment to 1 for armcc compiler */
 #define MP_PACKED    __packed
@@ -841,7 +842,7 @@ mp_encode_bin(char *data, const char *str, uint32_t len);
  * %s - zero-end string
  * %.*s - string with specified length
  * %% is ignored
- * %<smth else> assert and undefined behaviour
+ * %smthelse assert and undefined behaviour
  * NIL - a nil value
  * all other symbols are ignored.
  *
@@ -859,11 +860,22 @@ mp_format(char *data, size_t data_size, const char *format, ...);
  *  va_start(args, fmt);
  *  mp_vformat(data, data_size, fmt, args);
  *  va_end(args);
- * \sa \link mp_format()
+ * \sa \link mp_format() mp_format() \endlink
  */
 MP_PROTO size_t
 mp_vformat(char *data, size_t data_size, const char *format, va_list args);
 
+/**
+ * \brief print to \a file msgpacked data in JSON format.
+ * MP_EXT is printed as "EXT" only
+ * \param file - pointer to file (or NULL for stdout)
+ * \param data - pointer to buffer containing msgpack object
+ * \retval 0 - success
+ * \retval -1 - wrong msgpack
+ */
+MP_PROTO int
+mp_fprint(FILE* file, const char *data);
+
 /**
  * \brief Check that \a cur buffer has enough bytes to decode a string header
  * \param cur buffer
@@ -930,6 +942,27 @@ mp_decode_binl(const char **data);
 MP_PROTO const char *
 mp_decode_bin(const char **data, uint32_t *len);
 
+/**
+ * \brief Decode a length of a string or binstring from MsgPack \a data
+ * \param data - the pointer to a buffer
+ * \return a length of a string
+ * \post *data = *data + mp_sizeof_strbinl(retval)
+ * \sa mp_encode_binl
+ */
+MP_PROTO uint32_t
+mp_decode_strbinl(const char **data);
+
+/**
+ * \brief Decode a string or binstring from MsgPack \a data
+ * \param data - the pointer to a buffer
+ * \param len - the pointer to save a binstring length
+ * \return a pointer to a decoded binstring
+ * \post *data = *data + mp_sizeof_strbinl(*len)
+ * \sa mp_encode_binl
+ */
+MP_PROTO const char *
+mp_decode_strbin(const char **data, uint32_t *len);
+
 /**
  * \brief Calculate exact buffer size needed to store the nil value.
  * The return value is always 1. The function was added to provide integrity of
@@ -1094,9 +1127,10 @@ mp_check(const char **data, const char *end);
  * {{{ Implementation
  */
 
-/** \cond 0 */
+/** \cond false */
 extern const enum mp_type mp_type_hint[];
 extern const int8_t mp_parser_hint[];
+extern const char *mp_char2escape[];
 
 MP_IMPL MP_ALWAYSINLINE enum mp_type
 mp_typeof(const char c)
@@ -1222,14 +1256,14 @@ mp_decode_map(const char **data)
 {
 	uint8_t c = mp_load_u8(data);
 	switch (c) {
-	case 0x80 ... 0x8f:
-		return c & 0xf;
 	case 0xde:
 		return mp_load_u16(data);
 	case 0xdf:
 		return mp_load_u32(data);
 	default:
-		mp_unreachable();
+		if (mp_unlikely(c < 0x80 || c > 0x8f))
+			mp_unreachable();
+		return c & 0xf;
 	}
 }
 
@@ -1329,10 +1363,7 @@ MP_IMPL uint64_t
 mp_decode_uint(const char **data)
 {
 	uint8_t c = mp_load_u8(data);
-
 	switch (c) {
-	case 0x00 ... 0x7f:
-		return c;
 	case 0xcc:
 		return mp_load_u8(data);
 	case 0xcd:
@@ -1342,7 +1373,9 @@ mp_decode_uint(const char **data)
 	case 0xcf:
 		return mp_load_u64(data);
 	default:
-		mp_unreachable();
+		if (mp_unlikely(c > 0x7f))
+			mp_unreachable();
+		return c;
 	}
 }
 
@@ -1391,8 +1424,6 @@ mp_decode_int(const char **data)
 {
 	uint8_t c = mp_load_u8(data);
 	switch (c) {
-	case 0xe0 ... 0xff:
-		return (int8_t) (c);
 	case 0xd0:
 		return (int8_t) mp_load_u8(data);
 	case 0xd1:
@@ -1402,7 +1433,9 @@ mp_decode_int(const char **data)
 	case 0xd3:
 		return (int64_t) mp_load_u64(data);
 	default:
-		mp_unreachable();
+		if (mp_unlikely(c < 0xe0))
+			mp_unreachable();
+		return (int8_t) (c);
 	}
 }
 
@@ -1585,10 +1618,7 @@ MP_IMPL uint32_t
 mp_decode_strl(const char **data)
 {
 	uint8_t c = mp_load_u8(data);
-
 	switch (c) {
-	case 0xa0 ... 0xbf:
-		return c & 0x1f;
 	case 0xd9:
 		return mp_load_u8(data);
 	case 0xda:
@@ -1596,7 +1626,9 @@ mp_decode_strl(const char **data)
 	case 0xdb:
 		return mp_load_u32(data);
 	default:
-		mp_unreachable();
+		if (mp_unlikely(c < 0xa0 || c > 0xbf))
+			mp_unreachable();
+		return c & 0x1f;
 	}
 }
 
@@ -1639,6 +1671,42 @@ mp_decode_bin(const char **data, uint32_t *len)
 	return str;
 }
 
+MP_IMPL uint32_t
+mp_decode_strbinl(const char **data)
+{
+	uint8_t c = mp_load_u8(data);
+
+	switch (c) {
+	case 0xd9:
+		return mp_load_u8(data);
+	case 0xda:
+		return mp_load_u16(data);
+	case 0xdb:
+		return mp_load_u32(data);
+	case 0xc4:
+		return mp_load_u8(data);
+	case 0xc5:
+		return mp_load_u16(data);
+	case 0xc6:
+		return mp_load_u32(data);
+	default:
+		if (mp_unlikely(c < 0xa0 || c > 0xbf))
+			mp_unreachable();
+		return c & 0x1f;
+	}
+}
+
+MP_IMPL const char *
+mp_decode_strbin(const char **data, uint32_t *len)
+{
+	assert(len != NULL);
+
+	*len = mp_decode_strbinl(data);
+	const char *str = *data;
+	*data += *len;
+	return str;
+}
+
 MP_IMPL uint32_t
 mp_sizeof_nil()
 {
@@ -2095,6 +2163,104 @@ mp_format(char *data, size_t data_size, const char *format, ...)
 	return res;
 }
 
+MP_PROTO int
+mp_fprint_internal(FILE *file, const char **data);
+
+MP_IMPL int
+mp_fprint_internal(FILE *file, const char **data)
+{
+#define _CHECK_RC(exp) do { if (mp_unlikely((exp) < 0)) return -1; } while(0)
+	switch (mp_typeof(**data)) {
+	case MP_NIL:
+		mp_decode_nil(data);
+		_CHECK_RC(fputs("null", file));
+		break;
+	case MP_UINT:
+		_CHECK_RC(fprintf(file, "%llu", (unsigned long long)
+			      mp_decode_uint(data)));
+		break;
+	case MP_INT:
+		_CHECK_RC(fprintf(file, "%lld", (long long)
+			      mp_decode_int(data)));
+		break;
+	case MP_STR:
+	case MP_BIN:
+	{
+		uint32_t len = mp_typeof(**data) == MP_STR ?
+			mp_decode_strl(data) : mp_decode_binl(data);
+		_CHECK_RC(fputc('"', file));
+		const char *s;
+		for (s = *data; s < *data + len; s++) {
+			unsigned char c = (unsigned char ) *s;
+			if (c < 128 && mp_char2escape[c] != NULL) {
+				/* Escape character */
+				_CHECK_RC(fputs(mp_char2escape[c], file));
+			} else {
+				_CHECK_RC(fputc(c, file));
+			}
+		}
+		_CHECK_RC(fputc('"', file));
+		*data += len;
+		break;
+	}
+	case MP_ARRAY:
+	{
+		uint32_t size = mp_decode_array(data);
+		_CHECK_RC(fputc('[', file));
+		uint32_t i;
+		for (i = 0; i < size; i++) {
+			if (i)
+				_CHECK_RC(fputs(", ", file));
+			_CHECK_RC(mp_fprint_internal(file, data));
+		}
+		_CHECK_RC(fputc(']', file));
+		break;
+	}
+	case MP_MAP:
+	{
+		uint32_t size = mp_decode_map(data);
+		_CHECK_RC(fputc('{', file));
+		uint32_t i;
+		for (i = 0; i < size; i++) {
+			if (i)
+				_CHECK_RC(fprintf(file, ", "));
+			_CHECK_RC(mp_fprint_internal(file, data));
+			_CHECK_RC(fputs(": ", file));
+			_CHECK_RC(mp_fprint_internal(file, data));
+		}
+		_CHECK_RC(fputc('}', file));
+		break;
+	}
+	case MP_BOOL:
+		_CHECK_RC(fputs(mp_decode_bool(data) ? "true" : "false", file));
+		break;
+	case MP_FLOAT:
+		_CHECK_RC(fprintf(file, "%g", mp_decode_float(data)));
+		break;
+	case MP_DOUBLE:
+		_CHECK_RC(fprintf(file, "%lg", mp_decode_double(data)));
+		break;
+	case MP_EXT:
+		mp_next(data);
+		_CHECK_RC(fputs("undefined", file));
+		break;
+	default:
+		mp_unreachable();
+		return -1;
+	}
+	return 0;
+#undef _CHECK_RC
+}
+
+MP_IMPL int
+mp_fprint(FILE *file, const char *data)
+{
+	if (!file)
+		file = stdout;
+	int res = mp_fprint_internal(file, &data);
+	return res;
+}
+
 /** \endcond */
 
 /*
@@ -2105,7 +2271,7 @@ mp_format(char *data, size_t data_size, const char *format, ...)
  * {{{ Implementation: parser tables
  */
 
-/** \cond 0 */
+/** \cond false */
 
 #if defined(MP_SOURCE)
 
@@ -2727,6 +2893,29 @@ const int8_t mp_parser_hint[256] = {
 	/* }}} */
 };
 
+const char *mp_char2escape[128] = {
+	"\\u0000", "\\u0001", "\\u0002", "\\u0003",
+	"\\u0004", "\\u0005", "\\u0006", "\\u0007",
+	"\\b", "\\t", "\\n", "\\u000b",
+	"\\f", "\\r", "\\u000e", "\\u000f",
+	"\\u0010", "\\u0011", "\\u0012", "\\u0013",
+	"\\u0014", "\\u0015", "\\u0016", "\\u0017",
+	"\\u0018", "\\u0019", "\\u001a", "\\u001b",
+	"\\u001c", "\\u001d", "\\u001e", "\\u001f",
+	NULL, NULL, "\\\"", NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, "\\/",
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, "\\\\", NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
+	NULL, NULL, NULL, NULL, NULL, NULL, NULL, "\\u007f"
+};
+
 #endif /* defined(MP_SOURCE) */
 
 /** \endcond */
