@@ -38,10 +38,10 @@
 #include "msgpuck.h"
 #include "test.h"
 
-#define BUF_MAXLEN (1L << 18)
+#define BUF_MAXLEN ((1L << 18) - 1)
 #define STRBIN_MAXLEN (BUF_MAXLEN - 10)
 
-static char buf[BUF_MAXLEN];
+static char buf[BUF_MAXLEN + 1];
 static char str[STRBIN_MAXLEN];
 static char *data = buf + 1; /* use unaligned address to fail early */
 
@@ -77,10 +77,14 @@ static char *data = buf + 1; /* use unaligned address to fail early */
 		s1[i] = 'a' + i % 26;                                          \
 	}                                                                      \
 	const char *d1 = mp_encode_##_type(data, s1, _vl);                     \
-	const char *d2 = data;                                                 \
+	const char *d2;                                                        \
 	uint32_t len2;                                                         \
+	d2 = data;                                                             \
 	const char *s2 = mp_decode_##_type(&d2, &len2);                        \
 	is(_vl, len2, "len(mp_decode_"#_type"(x, %u))", _vl);                  \
+	d2 = data;                                                             \
+	(void) mp_decode_strbin(&d2, &len2);                                   \
+	is(_vl, len2, "len(mp_decode_strbin(x, %u))", _vl);                    \
 	const char *d3 = data;                                                 \
 	mp_next(&d3);                                                          \
 	const char *d4 = data;                                                 \
@@ -340,7 +344,7 @@ test_binls(void)
 static int
 test_strs(void)
 {
-	plan(84);
+	plan(96);
 	header();
 
 	test_str(0x01);
@@ -363,7 +367,7 @@ test_strs(void)
 static int
 test_bins(void)
 {
-	plan(84);
+	plan(96);
 	header();
 
 	test_bin(0x01);
@@ -396,10 +400,10 @@ test_next_on_array(uint32_t count)
 	const char *d2 = data;
 	const char *d3 = data;
 	ok(!mp_check(&d2, data + BUF_MAXLEN), "mp_check(array %u))", count);
-	is((d1 - data), len, "len(array %u) == %u", count, len);
-	is((d2 - data), len, "len(mp_check(array %u)) == %u", count, len);
+	is((d1 - data), (ptrdiff_t)len, "len(array %u) == %u", count, len);
+	is((d2 - data), (ptrdiff_t)len, "len(mp_check(array %u)) == %u", count, len);
 	mp_next(&d3);
-	is((d3 - data), len, "len(mp_next(array %u)) == %u", count, len);
+	is((d3 - data), (ptrdiff_t)len, "len(mp_next(array %u)) == %u", count, len);
 }
 
 static int
@@ -439,10 +443,10 @@ test_next_on_map(uint32_t count)
 	const char *d2 = data;
 	const char *d3 = data;
 	ok(!mp_check(&d2, data + BUF_MAXLEN), "mp_check(map %u))", count);
-	is((d1 - data), len, "len(map %u) == %u", count, len);
-	is((d2 - data), len, "len(mp_check(map %u)) == %u", count, len);
+	is((d1 - data), (ptrdiff_t)len, "len(map %u) == %u", count, len);
+	is((d2 - data), (ptrdiff_t)len, "len(mp_check(map %u)) == %u", count, len);
 	mp_next(&d3);
-	is((d3 - data), len, "len(mp_next(map %u)) == %u", count, len);
+	is((d3 - data), (ptrdiff_t)len, "len(mp_next(map %u)) == %u", count, len);
 }
 
 static int
@@ -716,9 +720,60 @@ test_format(void)
 	return check_plan();
 }
 
+int
+test_mp_print()
+{
+	plan(1);
+	header();
+
+	char data[512];
+
+	char *d = data;
+	d = mp_encode_array(d, 6);
+	d = mp_encode_int(d, -5);
+	d = mp_encode_uint(d, 42);
+	d = mp_encode_str(d, "kill bill", 9);
+	d = mp_encode_map(d, 6);
+	d = mp_encode_str(d, "bool true", 9);
+	d = mp_encode_bool(d, true);
+	d = mp_encode_str(d, "bool false", 10);
+	d = mp_encode_bool(d, false);
+	d = mp_encode_str(d, "null", 4);
+	d = mp_encode_nil(d);
+	d = mp_encode_str(d, "float", 5);
+	d = mp_encode_float(d, 3.14);
+	d = mp_encode_str(d, "double", 6);
+	d = mp_encode_double(d, 3.14);
+	d = mp_encode_uint(d, 100);
+	d = mp_encode_uint(d, 500);
+	*d++ = 0xd4; /* let's pack smallest fixed ext */
+	*d++ = 0;
+	*d++ = 0;
+	char bin[] = "\x12test\x34\b\t\n\"bla\\-bla\"\f\r";
+	d = mp_encode_bin(d, bin, sizeof(bin));
+
+	const char *expected =
+		"[-5, 42, \"kill bill\", "
+		"{\"bool true\": true, \"bool false\": false, \"null\": null, "
+		"\"float\": 3.14, \"double\": 3.14, 100: 500}, undefined, "
+		"\"\\u0012test4\\b\\t\\n\\\"bla\\\\-bla\\\"\\f\\r\\u0000\"]";
+	FILE *tmpf = tmpfile();
+	if (tmpf != NULL) {
+		mp_fprint(tmpf, data);
+		(void) rewind(tmpf);
+		memset(data, 0, sizeof(data));
+		if (fgets(data, sizeof(data), tmpf) != NULL) {
+			ok(strcmp(data, expected) == 0, "identical");
+		}
+		fclose(tmpf);
+	}
+	footer();
+	return check_plan();
+}
+
 int main()
 {
-	plan(16);
+	plan(17);
 
 	test_uints();
 	test_ints();
@@ -736,6 +791,7 @@ int main()
 	test_next_on_maps();
 	test_compare_uints();
 	test_format();
+	test_mp_print();
 
 	return check_plan();
 }
